VERSION 5.00
Begin VB.Form frmMain 
   BorderStyle     =   1  'Fixed Single
   Caption         =   "Convert Source to Wav"
   ClientHeight    =   1005
   ClientLeft      =   45
   ClientTop       =   375
   ClientWidth     =   4605
   BeginProperty Font 
      Name            =   "Courier New"
      Size            =   9.75
      Charset         =   0
      Weight          =   400
      Underline       =   0   'False
      Italic          =   0   'False
      Strikethrough   =   0   'False
   EndProperty
   LinkTopic       =   "Form1"
   MaxButton       =   0   'False
   MinButton       =   0   'False
   ScaleHeight     =   67
   ScaleMode       =   3  'Pixel
   ScaleWidth      =   307
   StartUpPosition =   3  'Windows Default
   Begin VB.PictureBox picPercent 
      AutoRedraw      =   -1  'True
      BeginProperty Font 
         Name            =   "Courier New"
         Size            =   8.25
         Charset         =   0
         Weight          =   400
         Underline       =   0   'False
         Italic          =   0   'False
         Strikethrough   =   0   'False
      EndProperty
      Height          =   300
      Left            =   45
      ScaleHeight     =   16
      ScaleMode       =   3  'Pixel
      ScaleWidth      =   296
      TabIndex        =   1
      Top             =   45
      Width           =   4500
   End
   Begin VB.CommandButton cmdConvert 
      Caption         =   "Convert to Wav"
      Height          =   540
      Left            =   45
      TabIndex        =   0
      Top             =   405
      Width           =   4500
   End
End
Attribute VB_Name = "frmMain"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = True
Attribute VB_Exposed = False
' Autor: F. Schler (frank@activevb.de)
' Datum: 08/2020
'
' Das ist nur ein Beispiel/Testcode! Fr entsprechende Infos zu den Interfaces, die
' hier im Beispiel verwendet werden, und wie das ganze funktioniert, bitte die
' Media Foundation Doku lesen!
'
' In diesem Beispiel wird nur das Prinzip gezeigt wie aus einer beliebigen, von der
' Media Foundation untersttzten, Audio- oder Video-Datei eine WAV-Audio Datei erstellt
' werden kann.
'
' Info zum Finalisieren (Function Finalize)
' https://docs.microsoft.com/en-us/windows/win32/medfound/media-sinks#finalizing
' Normalerweise nutzt man die Interfaces IMFFinalizableMediaSink und IMFAsyncResult
' um die fehlenden Daten nach dem letzten IMFSample (EndOfStream) zu schreiben.
' Die Verwendung dieser Interfaces in VBC fhrt, ohne ASM, zum Absturz der IDE bzw.
' des kompilierten Codes. Die Funktion Finalize macht im Prinzip das gleiche.
' Die Wave enthlt dann einen zustzlichen 'JUNK'-Chunk.
'
' Falls man einen Standard WavHeader haben mchte, kann man bei der Funktion
' ConvertSourceToWav -> bWriteStdWavHeader auf True setzen.

Option Explicit

Private Const Channels As Long = 2       ' 1, 2 usw.
Private Const BitsPerSample As Long = 16 ' 8, 16, 24, 32
Private Const SampleRate As Long = 44100 ' 6000, 8000, 11025, 16000, 22050, 32000, 44100,
                                         ' 48000, 64000, 88200, 96000, 176400, 192000

Private m_MFInit As Boolean

Private Sub Form_Load()
    
    picPercent.Scale (0, 0)-(100, 1)
    picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
       (picPercent.TextWidth("0%") / 2)
        
    picPercent.CurrentY = 0
    picPercent.ForeColor = vbBlack
    picPercent.Print "0%"
    
    ' Media Foundation starten
    If MFStartup(MF_VERSION, MF_STARTUP_FULL) = S_OK Then
        
        m_MFInit = True

    End If

End Sub

Private Sub cmdConvert_Click()

    Dim sSource As String
    Dim sWavFile As String
    
    ' sSource entsprechend anpassen!
    ' sWavFile entsprechend anpassen und die Extension sollte .wav sein!
    
    sSource = App.Path & "\test.mp3" ' diverse Audio- oder Video-Formate die
                                     ' von der Media Foundation untersttzt werden.
    
    sWavFile = App.Path & "\new.wav" ' zu erstellende Wav-Datei. Eine evtl. vorhandene
                                     ' Datei mit gleichem Namen wird automatisch gelscht!
                                     ' Siehe CreateWavWriter -> Flag MF_OPENMODE_DELETE_IF_EXIST
    
    ' ist die Media Foundation gestartet
    If m_MFInit = True Then
    
        ' Konvertiert sSource zu einer Wav-Datei (sWavFile)
        ' bWriteStdWavHeader:
        ' False (Defaut) = schreibt nur die fehlenden Werte in den vorhandenen WavHeader
        ' True = schreibt einen neuen Standard-WavHeader (44 Byte)
        If ConvertSourceToWav(sSource, sWavFile) = True Then
        
            MsgBox "Die Konvertierung ist abgeschlossen.", _
                vbOKOnly Or vbInformation, "Convert Source to Wav"
                
        Else
        
            MsgBox "Die Konvertierung ist fehlgeschlagen!", _
                vbOKOnly Or vbCritical, "Convert Source to Wav"
        
        End If
    
    End If
    
End Sub

Private Sub Form_Unload(Cancel As Integer)
    
    ' ist die Media Foundation gestartet
    If m_MFInit = True Then
    
        ' Media Foundation beenden
        Call MFShutdown
    
    End If

End Sub

' Konvertiert sInFile zu einer Wav-Datei (sOutFile)
Private Function ConvertSourceToWav(ByVal sInFile As String, ByVal sOutFile As _
    String, Optional ByVal bWriteStdWavHeader As Boolean = False) As Boolean
    
    Dim bRet As Boolean
    Dim bOutFile As Boolean
    Dim bUpDown As Boolean
    Dim bEndOfStream As Boolean
    Dim sPercent As Single
    Dim lPercent As Long
    Dim lStreamIndex As Long
    Dim lStreamFlags As Long
    Dim cTimeStamp As Currency
    Dim cSourceDuration As Currency
    Dim cSampleDuration As Currency
    Dim pIMFSample As Long
    Dim pIMFMediaSink As Long
    Dim pIMFStreamSink As Long
    Dim pIMFByteStream As Long
    Dim pIMFSourceReader As Long

    bRet = False
    bUpDown = False
    bOutFile = False
    bEndOfStream = False
    picPercent.ForeColor = vbBlack
    
    ' SourceReader von einer Datei erstellen
    If CreateSourceReader(sInFile, pIMFSourceReader) = True Then
    
        ' SourceDuration ermitteln
        If GetSourceDuration(pIMFSourceReader, cSourceDuration) = True Then
            
            picPercent.Cls
            picPercent.Scale (0, 0)-(100, 1)
                
        End If
    
        ' WavWriter von einer Datei erstellen
        If CreateWavWriter(sOutFile, pIMFByteStream, pIMFMediaSink, _
            pIMFStreamSink) = True Then
            
            bOutFile = True
            
            Do
            
                ' liet ein Sample vom SourceReader
                If Invoke(pIMFSourceReader, ReadSample, MF_SOURCE_READER_FIRST_AUDIO_STREAM, _
                    0&, VarPtr(lStreamIndex), VarPtr(lStreamFlags), VarPtr(cTimeStamp), _
                    VarPtr(pIMFSample)) = S_OK Then
                    
                    ' wenn nicht EndOfStream
                    If CBool(lStreamFlags And MF_SOURCE_READER_FLAG.MF_SOURCE_READERF_ENDOFSTREAM) = False Then
                        
                        ' nur wenn SourceDuration > 0
                        If cSourceDuration > 0@ Then
                        
                            ' SampleDuration ermitteln
                            If GetSampleDuration(pIMFSample, cSampleDuration) = True Then
                                
                                ' Zeiten in Prozent umrechnen
                                lPercent = ((cTimeStamp + cSampleDuration) * 100) \ cSourceDuration
                                
                                ' Fortschritt ausgeben
                                picPercent.Line (0, 0)-(lPercent, 1), vbGreen, BF
                                
                                picPercent.Line (lPercent, 0)-(100, 1), _
                                    picPercent.BackColor, BF
                                
                                picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
                                    (picPercent.TextWidth(CStr(lPercent) & "%") / 2)
        
                                picPercent.CurrentY = 0
                                picPercent.Print CStr(lPercent) & "%"

                                
                            End If
                            
                        Else
                            
                            ' Anzeige wenn SourceDuration = 0
                            If sPercent = 50 Then
                            
                                sPercent = 0
                                bUpDown = Not bUpDown
                                
                            End If
                            
                            If bUpDown = False Then
                            
                                picPercent.Line (50 - sPercent, 0)-(50 + sPercent, 1), _
                                    vbGreen, BF
                                
                            Else
                            
                                picPercent.Line (50 - sPercent, 0)-(50 + sPercent, 1), _
                                    picPercent.BackColor, BF
                            
                            End If
                            
                            picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
                                (picPercent.TextWidth("Konvertierung luft...") / 2)
        
                            picPercent.CurrentY = 0
                            picPercent.Print "Konvertierung luft..."
                            
                            sPercent = sPercent + 0.05
                            
                        End If
                        
                        ' schreibt ein Sample in den StreamSink
                        If Invoke(pIMFStreamSink, ProcessSample, pIMFSample) <> S_OK Then
                            
                            ' bei einem Fehler abbrechen
                            bEndOfStream = True
                            
                        End If
                        
                    End If
                    
                    ' wenn EndOfStream
                    If CBool(lStreamFlags And MF_SOURCE_READER_FLAG.MF_SOURCE_READERF_ENDOFSTREAM) = True Then
                    
                        ' EndOfStream erreicht
                        bEndOfStream = True
                        
                        ' Rest in den StreamSink schreiben
                        If Invoke(pIMFStreamSink, Flush_SS) = S_OK Then
                        
                            bRet = True
                            picPercent.Line (0, 0)-(100, 1), vbGreen, BF
                
                            picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
                                (picPercent.TextWidth("100%") / 2)
        
                            picPercent.CurrentY = 0
                            picPercent.Print "100%"
                
                        End If
                        
                    End If
                    
                    ' freigeben
                    Call Release(pIMFSample)
                    
                End If
                
                DoEvents
                
            ' bei Fehler oder EndOfStream die Schleife verlassen
            Loop Until bEndOfStream = True
            
            ' freigeben
            Call Release(pIMFStreamSink)
            
            ' Rest in den ByteStream schreiben
            If Invoke(pIMFByteStream, Flush_BS) = S_OK Then
            
                ' wenn kein Standard-WavHeader geschrieben werden soll
                If bWriteStdWavHeader = False Then
                    
                    ' fehlende Werte schreiben
                    bRet = Finalize(pIMFByteStream)
                    
                End If
            
                ' ByteStream schlieen
                If Invoke(pIMFByteStream, Close_BS) = S_OK Then

                    ' freigeben
                    Call Release(pIMFByteStream)

                End If

            End If
            
            ' MediaSink schlieen
            If Invoke(pIMFMediaSink, Shutdown) = S_OK Then
                
                ' freigeben
                Call Release(pIMFMediaSink)

            End If
            
        End If
        
        ' freigeben
        Call Release(pIMFSourceReader)
        
    End If
    
    ' wenn ein Standard-WavHeader geschrieben werden soll
    If bWriteStdWavHeader = True Then
    
        If bRet = True Then
        
            ' schreibe einen Standard WavHeader
            bRet = WriteNewWaveHeader(sOutFile)
            
        End If
        
    End If
    
    If bRet = False Then
        
        picPercent.Line (0, 0)-(picPercent.ScaleWidth, _
            picPercent.ScaleHeight), vbRed, BF
    
        picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
            (picPercent.TextWidth("Ups, da ist was schiefgelaufen. :-(") / 2)
        
        picPercent.CurrentY = 0
        picPercent.ForeColor = vbWhite
        picPercent.Print "Ups, da ist was schiefgelaufen. :-("
    
        DoEvents
        
        ' im Fehlerfall kann die erzeugte Datei weg
        If bOutFile = True Then Call Kill(sOutFile)
        
    End If
    
    ConvertSourceToWav = bRet
    
End Function

' SourceReader von einer Datei erstellen
Private Function CreateSourceReader(ByVal sFile As String, _
    ByRef pReader As Long) As Boolean

    Dim bRet As Boolean
    Dim pIMFMediaType As Long

    bRet = False

    ' SourceReader erstellen
    If MFCreateSourceReaderFromURL(StrPtr(sFile), 0&, pReader) = S_OK Then

        ' MediaType fr den SourceReader erstellen/festlegen
        If CreatePCMAudioType(SampleRate, BitsPerSample, _
            Channels, pIMFMediaType) = True Then

            ' MediaType fr den SourceReader setzen
            If Invoke(pReader, SetCurrentMediaType, _
                MF_SOURCE_READER_FIRST_AUDIO_STREAM, _
                0&, pIMFMediaType) = S_OK Then

                bRet = True

            End If

            ' freigeben
            Call Release(pIMFMediaType)

        End If

    End If

    CreateSourceReader = bRet

End Function

' WavWriter von einer Datei erstellen
Private Function CreateWavWriter(ByVal sFile As String, _
    ByRef pByteStream As Long, ByRef pMediaSink As Long, _
    ByRef pStreamSink As Long) As Boolean
    
    Dim bRet As Boolean
    Dim lSinkCount As Long
    Dim pIMFMediaType As Long

    bRet = False
    
    ' ByteStream von einer Datei erstellen.
    ' Eine vorhandene Datei mit gleichem Namen
    ' wird gelscht! -> MF_OPENMODE_DELETE_IF_EXIST
    If MFCreateFile(MF_ACCESSMODE_READWRITE, MF_OPENMODE_DELETE_IF_EXIST, _
        MF_FILEFLAGS_NONE, StrPtr(sFile), pByteStream) = S_OK Then
        
        ' MediaType fr den WavWriter erstellen/festlegen
        If CreatePCMAudioType(SampleRate, BitsPerSample, _
            Channels, pIMFMediaType) = True Then

            ' WAVEMediaSink erstellen
            If MFCreateWAVEMediaSink(pByteStream, pIMFMediaType, _
                pMediaSink) = S_OK Then
            
                ' anzahl der StreamSinks ermitteln
                If Invoke(pMediaSink, GetStreamSinkCount, _
                    VarPtr(lSinkCount)) = S_OK Then
                
                    ' ist mindestens ein StreamSink vorhanden
                    If lSinkCount > 0& Then
                    
                        ' ersten StreamSink erstellen
                        If Invoke(pMediaSink, GetStreamSinkByIndex, 0&, _
                            VarPtr(pStreamSink)) = S_OK Then
                        
                            bRet = True
                        
                        End If
                    
                    End If
                
                End If
            
            End If

            ' freigeben
            Call Release(pIMFMediaType)

        End If
        
    End If
    
    CreateWavWriter = bRet
    
End Function

' MediaType fr den SourceReader und dem SinkWriter erstellen
Private Function CreatePCMAudioType(ByVal lSampleRate As Long, _
    ByVal lBitsPerSample As Long, ByVal lChannels As Long, _
    ByRef pType As Long) As Boolean

    Dim bRet As Boolean
    Dim lBlockAlign As Long
    Dim lBytesPerSecond As Long
    
    bRet = False

    lBlockAlign = lChannels * (lBitsPerSample / 8)
    lBytesPerSecond = lBlockAlign * lSampleRate
    
    ' MediaType erstellen
    If MFCreateMediaType(pType) = S_OK Then
        
        ' MajorType Audio
        If Invoke(pType, SetGUID, VarPtr(Str2Guid(MF_MT_MAJOR_TYPE)), _
            VarPtr(Str2Guid(MFMediaType_Audio))) = S_OK Then
            
            ' SubType PCM
            If Invoke(pType, SetGUID, VarPtr(Str2Guid(MF_MT_SUBTYPE)), _
                VarPtr(Str2Guid(MFAudioFormat_PCM))) = S_OK Then
            
                If Invoke(pType, SetUINT32, _
                    VarPtr(Str2Guid(MF_MT_AUDIO_SAMPLES_PER_SECOND)), _
                    lSampleRate) = S_OK Then
                    
                    If Invoke(pType, SetUINT32, _
                        VarPtr(Str2Guid(MF_MT_AUDIO_BITS_PER_SAMPLE)), _
                        lBitsPerSample) = S_OK Then

                        If Invoke(pType, SetUINT32, _
                            VarPtr(Str2Guid(MF_MT_AUDIO_NUM_CHANNELS)), _
                            lChannels) = S_OK Then
                        
                            If Invoke(pType, SetUINT32, _
                                VarPtr(Str2Guid(MF_MT_AUDIO_BLOCK_ALIGNMENT)), _
                                lBlockAlign) = S_OK Then
                        
                                If Invoke(pType, SetUINT32, _
                                    VarPtr(Str2Guid(MF_MT_AUDIO_AVG_BYTES_PER_SECOND)), _
                                    lBytesPerSecond) = S_OK Then
                        
                                    If Invoke(pType, SetUINT32, _
                                        VarPtr(Str2Guid(MF_MT_ALL_SAMPLES_INDEPENDENT)), _
                                        API_True) = S_OK Then
                        
                                        bRet = True
                                        
                                    End If
                                
                                End If
                                
                            End If
                            
                        End If
                        
                    End If
                    
                End If
                        
            End If
                
        End If
        
    End If

    CreatePCMAudioType = bRet
    
End Function

' SourceDuration ermitteln
Private Function GetSourceDuration(ByVal pReader As Long, _
    ByRef cDuration As Currency) As Boolean
    
    Dim bRet As Boolean
    Dim tPROPVARIANT As PROPVARIANT
    
    bRet = False
    
    ' SourceDuration ermitteln
    If Invoke(pReader, GetPresentationAttribute, _
        MF_SOURCE_READER_MEDIASOURCE, _
        VarPtr(Str2Guid(MF_PD_DURATION)), _
        VarPtr(tPROPVARIANT)) = S_OK Then
        
        ' PROPVARIANT zu Currency konvertieren
        If PropVariantToInt64(tPROPVARIANT, cDuration) = S_OK Then
        
            bRet = True
            
        End If
    
        ' PROPVARIANT freigeben
        Call PropVariantClear(tPROPVARIANT)
        
    End If

    GetSourceDuration = bRet
    
End Function

' SampleDuration ermitteln
Private Function GetSampleDuration(ByVal pSample As Long, _
    ByRef cDuration As Currency) As Boolean
    
    Dim bRet As Boolean
    
    bRet = False
    
    ' SampleDuration ermitteln
    If Invoke(pSample, GetSampleDuration_S, VarPtr(cDuration)) = S_OK Then
            
        bRet = True
            
    End If

    GetSampleDuration = bRet
    
End Function

' Entfernt den vorhandenen WaveHeader und ersetzt diesen durch
' einen neuen WaveHeader.
Private Function WriteNewWaveHeader(ByVal sWavFile As String) As Boolean

    Dim bRet As Boolean
    Dim bData() As Byte
    Dim hFile As Long
    Dim lRead As Long
    Dim lWritten As Long
    Dim lCutSize As Long
    Dim lFileSize As Long
    Dim lJunkSize As Long
    Dim lFmtSize As Long
    Dim cRead As Currency
    Dim cWritten As Currency
    Dim cNewPosition As Currency
    Dim pInIStream As Long
    Dim pOutIStream As Long
    Dim tWaveHeader As WaveHeader

    bRet = False
    
    picPercent.Line (0, 0)-(100, 1), vbGreen, BF
    
    picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
        (picPercent.TextWidth("Schreibe neuen WaveHeader") / 2)
        
    picPercent.CurrentY = 0
    picPercent.ForeColor = vbBlack
    picPercent.Print "Schreibe neuen WaveHeader"
    DoEvents

    ' Datei ffnen
    If SHCreateStreamOnFileEx(StrPtr(sWavFile), STGM_READWRITE Or _
        STGM_SHARE_DENY_NONE, 0&, 0&, 0&, pInIStream) = S_OK Then
        
        ' Datei nochmal ffnen (fr das umkopieren)
        If SHCreateStreamOnFileEx(StrPtr(sWavFile), STGM_WRITE Or _
            STGM_SHARE_DENY_NONE, 0&, 0&, 0&, pOutIStream) = S_OK Then
            
            ' Gre der Datei ermitteln
            If Invoke(pInIStream, Seek_Strm, CCur(0), STREAM_SEEK.STREAM_SEEK_END, _
                VarPtr(cNewPosition)) = S_OK Then
            
                lFileSize = Cur2Lng(cNewPosition)
                
                ' 4 Byte 'RIFF' + 4 Byte RiffSize + 4 Byte 'WAVE' + 4 Byte 'JUNK'
                lCutSize = 16
                
                ' zur angegebenen Postion gehen
                If Invoke(pInIStream, Seek_Strm, Lng2Cur(lCutSize), _
                    STREAM_SEEK.STREAM_SEEK_SET, VarPtr(cNewPosition)) = S_OK Then
                    
                    ReDim bData(3)
                    
                    ' Gre des 'JUNK'-Chunk ermitteln
                    If Invoke(pInIStream, Read_Strm, VarPtr(bData(0)), _
                        CLng(UBound(bData) + 1), VarPtr(lRead)) = S_OK Then
                    
                        Call CopyMemory(lJunkSize, bData(0), CLng(UBound(bData) + 1))
                        
                        ' lCutSize + 4 Byte JunkSize, lJunkSize, 4 Byte 'fmt ')
                        lCutSize = lCutSize + lJunkSize + 8
                        
                        ' zur angegebenen Postion gehen
                        If Invoke(pInIStream, Seek_Strm, Lng2Cur(lCutSize), _
                            STREAM_SEEK.STREAM_SEEK_SET, VarPtr(cNewPosition)) = S_OK Then
                        
                            ' Gre des 'fmt '-Chunk ermitteln
                            If Invoke(pInIStream, Read_Strm, VarPtr(bData(0)), _
                                CLng(UBound(bData) + 1), VarPtr(lRead)) = S_OK Then
                            
                                Call CopyMemory(lFmtSize, bData(0), CLng(UBound(bData) + 1))
                                
                                'lCutSize + 4 Byte fmtSize, lFmtSize, 4 Byte 'DATA', 4 byte DataSize)
                                lCutSize = lCutSize + lFmtSize + 12
                                
                                ' ab Position, wo die PCM-Daten beginnen, einlesen
                                If Invoke(pInIStream, Seek_Strm, Lng2Cur(lCutSize), _
                                    STREAM_SEEK.STREAM_SEEK_SET, VarPtr(cNewPosition)) = S_OK Then
                                    
                                    ' nach der Position vom WaveHeader schreiben
                                    If Invoke(pOutIStream, Seek_Strm, Lng2Cur(Len(tWaveHeader)), _
                                        STREAM_SEEK.STREAM_SEEK_SET, VarPtr(cNewPosition)) = S_OK Then
                                    
                                        ' PCM-Daten umkopieren
                                        If Invoke(pInIStream, CopyTo_Strm, pOutIStream, _
                                            Lng2Cur(lFileSize - lCutSize), _
                                            VarPtr(cRead), VarPtr(cWritten)) = S_OK Then
                                            
                                            ' zum Anfang gehen
                                            If Invoke(pInIStream, Seek_Strm, CCur(0), _
                                                STREAM_SEEK.STREAM_SEEK_SET, _
                                                VarPtr(cNewPosition)) = S_OK Then
                                            
                                                ' Waveheader festlegen
                                                With tWaveHeader
                                                    .lRiff = 1179011410
                                                    .lFileSize = (lFileSize - lCutSize) - 8
                                                    .lWave = 1163280727
                                                    .lFmt = 544501094
                                                    .lFmtSize = 16
                                                    .iFormatTag = 1
                                                    .iChannels = Channels
                                                    .lSamplesPerSec = SampleRate
                                                    .lAvgBytesPerSec = (BitsPerSample * SampleRate * Channels) / 8
                                                    .iBlockAlign = (BitsPerSample * Channels) / 8
                                                    .iBitsPerSample = BitsPerSample
                                                    .lData = 1635017060
                                                    .lDataSize = (lFileSize - lCutSize) - 44
                                                End With
                        
                                                ReDim bData(Len(tWaveHeader) - 1)
                                                
                                                Call CopyMemory(bData(0), tWaveHeader, CLng(UBound(bData) + 1))
                        
                                                ' Waveheader an den Anfang schreiben
                                                If Invoke(pInIStream, Write_Strm, VarPtr(bData(0)), _
                                                    CLng(UBound(bData) + 1), VarPtr(lWritten)) = S_OK Then
                                    
                                                    bRet = True
                                                    
                                                End If
                        
                                            End If
                                        
                                        End If
                                    
                                    End If
                                
                                End If
                                
                            End If
                        
                        End If
                        
                    End If
                
                End If
            
            End If
            
            ' freigeben
            Call Release(pOutIStream)
            
        End If
        
        ' freigeben
        Call Release(pInIStream)
            
    End If
    
    If bRet = True Then

        ' Datei ffnen
        hFile = CreateFile(StrPtr(sWavFile), GENERIC_WRITE, FILE_SHARE_READ, _
            0&, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0&)

        ' ist ein Handle auf die Datei vorhanden
        If hFile <> -1 Then

            ' Dateizeiger an die gewnschte Position setzen
            If SetFilePointer(hFile, -lCutSize, 0&, FILE_END) <> _
                INVALID_SET_FILE_POINTER Then

                ' Aktuelle Position als "END-Position" setzen
                If SetEndOfFile(hFile) = API_True Then
                
                    bRet = True
                    
                Else
                
                    bRet = False
                
                
                End If
                
            Else
            
                bRet = False
                
            End If

            ' Datei schlieen
            Call CloseHandle(hFile)

        Else

            bRet = False

        End If

    End If
    
    If bRet = True Then

        picPercent.Line (0, 0)-(100, 1), vbGreen, BF
    
        picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
            (picPercent.TextWidth("Fertig ;-)") / 2)
        
        picPercent.CurrentY = 0
        picPercent.ForeColor = vbBlack
        picPercent.Print "Fertig ;-)"
    
        DoEvents
    
    Else

        picPercent.Line (0, 0)-(picPercent.ScaleWidth, _
            picPercent.ScaleHeight), vbRed, BF
        
        picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
            (picPercent.TextWidth("Ups, da ist was schiefgelaufen. :-(") / 2)
        
        picPercent.CurrentY = 0
        picPercent.ForeColor = vbWhite
        picPercent.Print "Ups, da ist was schiefgelaufen. :-("
    
        DoEvents
    
    End If

    WriteNewWaveHeader = bRet
    
End Function

' schreibt fehlende Werte in den ByteStream
' Ersatz fr die Interfaces IMFFinalizableMediaSink und IMFAsyncResult
' siehe Info auf:
' https://docs.microsoft.com/en-us/windows/win32/medfound/media-sinks#finalizing
Private Function Finalize(ByVal pByteStream As Long) As Boolean
    
    Dim bRet As Boolean
    Dim bData() As Byte
    Dim lPos As Long
    Dim lDataPos As Long
    Dim lRead As Long
    Dim lLength As Long
    Dim lWritten As Long
    Dim lFmtSize As Long
    Dim lJunkSize As Long
    Dim cLength As Currency
    
    bRet = False
    
    picPercent.Line (0, 0)-(100, 1), vbGreen, BF
    
    picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
        (picPercent.TextWidth("Finalisieren") / 2)
        
    picPercent.CurrentY = 0
    picPercent.ForeColor = vbBlack
    picPercent.Print "Finalisieren"
    
    DoEvents
    
    ' Lnge des Streams ermitteln
    If Invoke(pByteStream, GetLength_BS, VarPtr(cLength)) = S_OK Then
    
        lLength = Cur2Lng(cLength)
        
        ' Position 'RIFF' (4 Byte) + 4 Byte Size + 'WAVE' (4 Byte) + 'JUNK' (4 Byte)
        lPos = 16
        
        If Invoke(pByteStream, SetCurrentPosition_BS, Lng2Cur(lPos)) = S_OK Then
        
            ReDim bData(3)
            
            ' Gre des JUNK-Chunk auslesen
            If Invoke(pByteStream, Read_BS, VarPtr(bData(0)), _
                UBound(bData) + 1, VarPtr(lRead)) = S_OK Then
            
                Call CopyMemory(lJunkSize, bData(0), UBound(bData) + 1)
                
                ' Position + 4 Byte JunkSize, lJunkSize, 4 Byte 'fmt ')
                lPos = lPos + lJunkSize + 8
                
                If Invoke(pByteStream, SetCurrentPosition_BS, Lng2Cur(lPos)) = S_OK Then
                
                    ' Gre des FMT-Chunk auslesen
                    If Invoke(pByteStream, Read_BS, VarPtr(bData(0)), _
                        UBound(bData) + 1, VarPtr(lRead)) = S_OK Then
                
                        Call CopyMemory(lFmtSize, bData(0), UBound(bData) + 1)
                        
                        'Position + 4 Byte fmtSize, lFmtSize, 4 Byte 'DATA')
                        lDataPos = lPos + lFmtSize + 8
                        
                        ' Position hinter 'RIFF'
                        If Invoke(pByteStream, SetCurrentPosition_BS, Lng2Cur(4)) = S_OK Then

                            ' Gre des Streams - 8 Byte
                            Call CopyMemory(bData(0), lLength - 8, UBound(bData) + 1)

                            ' Wert hinter 'RIFF' schreiben
                            If Invoke(pByteStream, Write_BS, VarPtr(bData(0)), _
                                UBound(bData) + 1, VarPtr(lWritten)) = S_OK Then
                            
                                ' Position hinter 'DATA'
                                If Invoke(pByteStream, SetCurrentPosition_BS, _
                                    Lng2Cur(lDataPos)) = S_OK Then
                                    
                                    ' Gre der PCM-Daten = Gre des Streams - (Position 'DATA' + 4 Byte DataSize)
                                    Call CopyMemory(bData(0), lLength - (lDataPos + 4), _
                                        UBound(bData) + 1)
                                
                                    ' Wert hinter 'DATA' schreiben
                                    If Invoke(pByteStream, Write_BS, VarPtr(bData(0)), _
                                        UBound(bData) + 1, VarPtr(lWritten)) = S_OK Then
                                    
                                        bRet = True
                                    
                                    End If
                                
                                End If

                            End If

                        End If
                
                    End If
                    
                End If
            
            End If
        
        End If
    
    End If

    If bRet = True Then

        picPercent.Line (0, 0)-(100, 1), vbGreen, BF
    
        picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
            (picPercent.TextWidth("Fertig ;-)") / 2)
        
        picPercent.CurrentY = 0
        picPercent.ForeColor = vbBlack
        picPercent.Print "Fertig ;-)"
    
        DoEvents
    
    Else

        picPercent.Line (0, 0)-(picPercent.ScaleWidth, _
            picPercent.ScaleHeight), vbRed, BF
        
        picPercent.CurrentX = (picPercent.ScaleWidth / 2) - _
            (picPercent.TextWidth("Ups, da ist was schiefgelaufen. :-(") / 2)
        
        picPercent.CurrentY = 0
        picPercent.ForeColor = vbWhite
        picPercent.Print "Ups, da ist was schiefgelaufen. :-("
    
        DoEvents
    
    End If

    Finalize = bRet
    
End Function

' Long -> Currency
Private Function Lng2Cur(ByVal lVal As Long) As Currency
    
    Lng2Cur = CCur(lVal) / 10000@

End Function

' Currency -> Long
Private Function Cur2Lng(ByVal lVal As Currency) As Long
    
    Cur2Lng = CLng(lVal * 10000@)

End Function

